#version 330 core

layout(location = 0) in vec3 vertexPosition;
layout(location = 1) in vec2 vertexTexCoord;
layout(location = 2) in vec3 vertexNormal;
layout(location = 3) in vec4 vertexColor;

out vec2 texCoord;
out vec4 vertexFragColor;
uniform mat4 mvp;

uniform sampler2D terrain;
uniform float time;

uniform float heightMagnitude = 1;
uniform float ySpace = 0.0;
uniform float xSpace = 0.04;
uniform float gridStep = 0.04;
uniform float timeMultiplier = 0.2;
out float xSpaceFrag;

void main() {
    xSpaceFrag = xSpace;
    vec2 texcoord = vertexTexCoord;
    texcoord.y = mod(texcoord.y + time * timeMultiplier, 1.0);
    texCoord = texcoord;


    float height = texture(terrain, texcoord).r * heightMagnitude;

    vec3 vert = vertexPosition;
    vert.z = height;

    vertexFragColor = vec4(0);

    float centerLine= 0.5;
    float horizontalCenter = abs(vertexTexCoord.x - centerLine);
    if (abs(vertexTexCoord.x - centerLine) < xSpaceFrag || abs(vertexTexCoord.y - centerLine) < ySpace) {
        vert.z = 0;
    } else {
        //griddify
        float step = gridStep;
        float xc = mod(vertexTexCoord.x, step)/step;
        float yc = mod(vertexTexCoord.y, step)/step;
        float cc = 1.0-clamp(min(xc,yc)*10,0.0,1.0);

        vertexFragColor = vec4(vec3(cc),1);
    }

    gl_Position =  mvp * vec4(vert, 1);

    //if (vert.z > 0.0) {
    //}
}
